/* timecli.c

   An example of using a UDP socket to obtain the current time
   from a remote server, based on RFC 868.

 */

#include "config.h"

#ifndef MSG_WAITALL
#define MSG_WAITALL 0
#endif

#ifndef HAS_int32_t
#define uint32_t unsigned int
#define int32_t unsigned long
#endif

#define CHECK(CALL)\
 if ((CALL) < 0) {\
    perror (#CALL);\
    _exit (-1);\
 }

#define ERR_QUIT(MSG) {\
 perror (MSG);\
 _exit (-1);\
 }

#define SA struct sockaddr

int
main(int argc, char **argv)
{
   int sockfd;
   struct sockaddr_in servaddr;
   struct hostent *hosts;
   char *server;
   struct tm *t;
   time_t time;
   int nsent, nrcvd;
   socklen_t addrlen;
   uint32_t buf, epoch;

   if (argc != 2) ERR_QUIT("usage: udpcli <IPaddress>");

   server = argv[1];
   if ( !(hosts = gethostbyname (server))) {
      ERR_QUIT ("gethostbyname");
   }

   memset (&servaddr, 0, sizeof(servaddr));
   servaddr.sin_family = AF_INET;
   servaddr.sin_port = htons(37);
   servaddr.sin_addr = *(struct in_addr *) (hosts->h_addr_list[0]);

   sockfd = socket(AF_INET, SOCK_DGRAM, 0);
   CHECK ( (nsent = sendto (sockfd, (void *) &buf, 0,
      0, (SA *) &servaddr, sizeof (servaddr))));
   addrlen = sizeof (servaddr);

   CHECK ( (nrcvd = recvfrom (sockfd, (void *) &buf, sizeof (buf), 
      MSG_WAITALL, (SA *) &servaddr, &addrlen)));

   /* the value received is a count of seconds since
      00:00:00 GMT, January 1, 1900

      We presume it is in network byte order.
    */

   buf = ntohl (buf);

   /* gmtime splits out a time that is measured in seconds since
      00:00:00 UTC, January 1, 1970 (the Epoch)
 
      To convert the value in buf to this frame of reference we
      need to subtract the number of seconds between
      00:00:00, January 1, 1900 and the Epoch.
      We can't use type time_t, since that cannot be used to
      represent times prior to the Epoch.
      How many seconds are there between these two dates?
      We have 1970 - 1900 = 70 years,
      but how about leap years?
      There were leap years every 4 years during this interval,
      except for 1900, so there should be 70/4 - 1 = 17 leap years.
      UTC also has leap seconds, but it seems these are not
      implemented for the cases I have tested.
    */

   epoch = ((unsigned long) (365) * 70 + 17) * 24 * 60 * 60;
   time = (time_t) (buf - epoch);
   t = gmtime (&time);
   fprintf (stdout, "time= %.2d:%.2d:%.2d %.2d/%.2d/%.2d\n",
     t->tm_hour, t->tm_min, t->tm_sec, 
     1900 + t->tm_year, t->tm_mon + 1, t->tm_mday);

   exit(0);
}







