/* philos.c
   @(#)philos.c	1.1 07 Jan 1994

   Main program, for Dining Philosophers exercise

   The effect of this program is to read an integer from the
   standard input, indicating an upper bound on how many trace-steps
   to allow before starting to shut down.  It will allow this many
   steps, and then shut down.  The shutdown will take a few more
   steps.  While the program is running, it will output a series
   of lines.  Each line begins with a number identifying the active
   philosopher, followed by a message.

   Prepared for COP4610, Fall 1993
   --Ted Baker

   Corrected error in shutdown mechanism, which could create
   deadlock of thread exits while holding a mutex, or chopsticks.
   --Ted Baker, 17 September 1993

   Corrected failure to initalize mutex "outlock".
   --Ted Baker, 29 Septempber 1993

 */

#include <stdio.h>
#include <ctype.h>
#include <pthread.h>
#include "chopsticks.h"

#define N 5                /* the number of philosophers */

pthread_mutex_t outlock;   /* protects against output interleaving */
int nsteps, maxsteps = 0;  /* number of steps to run this test */

int eat_count[N];          /* number of steps for each thread */

/* The subprogram trace is for testing and debugging.
   It allows us to see what is happening,
   and to shut down the philosophers when the program has
   run long enough for us to see that it seems to be working 
 */

void trace(int i, char *s)
  /* print out message, for use in execution tracing
     i = philospher ID
     s = message 
   */
{
  pthread_mutex_lock(&outlock);
  if (strcmp (s, "eating") == 0) eat_count [i]++;
  fprintf(stdout,"%d: %s\n",i,s);
  if (nsteps++ > maxsteps) {
    /* don't exit while we are holding any chopsticks */
    if (strcmp(s,"thinking") == 0) {
      pthread_mutex_unlock(&outlock); 
      fprintf (stderr, "thread done\n");
      pthread_exit(0);
    }
  }
  pthread_mutex_unlock(&outlock);
}

void * philosopher(void *arg)
{ int self = *(int *) arg;
  for (;;) {
     trace(self,"thinking");
     chopsticks_take(self);
     trace(self,"eating");
     chopsticks_put(self);
  }
}

main()
{
  int i; 
  pthread_t th[N]; /* IDs of the philospher threads */
  int       no[N]; /* corresponding table position numbers*/
  pthread_attr_t attr;

  for (i = 0; i < N; i++) eat_count [i] = 0;

  pthread_mutex_init(&outlock, NULL);

  /* initialize the package "chopsticks" */
  chopsticks_init();

  fprintf(stdout,"enter number of steps to run: "); fflush(stdout);
  fscanf(stdin,"%d",&maxsteps);

  pthread_attr_init (&attr); /* sets default attributes */

  thr_setconcurrency (4);
  /* Solaris-specific: suggest four LWP's for this process */

  pthread_attr_setscope (&attr, PTHREAD_SCOPE_SYSTEM);
  /* set system-wide contention scope */

  /* start up the philosopher threads */
  for (i = 0; i < N; i++) {
    no[i] = i;
    pthread_create(&th[i], NULL, philosopher, (int *) &no[i]);
  }

  /* wait for all the threads to shut down */
  for (i = 0; i < N; i++) pthread_join(th[i], NULL);

  for (i = 0; i < N; i++) {
     fprintf (stderr, "philospher %d ate %d times\n", i, eat_count [i]);
  }

  return 0;
}

/* (See file "prodcons.c" for more comments on use of POSIX
    threads operations.)
 */



